{% extends "base.html" %}
{% block title %}Recovery code · tit{% endblock %}
{% block content %}
  <h1>Save your recovery code</h1>
  <p>This code is shown one time. Store it offline. It replaces all SSH keys when you use it.</p>
  <pre><code>{{ recovery }}</code></pre>
{% endblock %}
